% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% Ph.D. Course "Intelligence for Embedded Systems"
% Prof. Manuel Roveri
% Politecnico di Milano, Italy
% Jan, 2015
% 
% ----- "Uncertainty, Information and Learning Mechanisms - Part 1"
% -> Example 3_2_E) Different numbers of hidden units
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


clear;
clc;
close all;

%% Data generation
%Number of points used for training
N = 50;

% Number of experiments for each configuration of the hidden neurons
n_repe = 50;

% Number of hidden nodes to explore
min_hidden_nodes = 2;
max_hidden_nodes = 20;

%Noise standard deviation
std_dev = 0.05;

%% MAIN %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Noisy data
Xt = 4 * rand(1,N) - 2;
NoiseT = std_dev * randn(1,N);
yt = nonlin_fun(Xt,NoiseT);

%% Network training

% Results initialization
n_experim = 1;


tot_experim = max_hidden_nodes - min_hidden_nodes + 1;
vali_error = zeros(tot_experim,1);
trai_error = zeros(tot_experim,1);
test_error = zeros(tot_experim,1);

for n_hidden_nodes = min_hidden_nodes:max_hidden_nodes

    spec_vali_error = zeros(n_repe,1);
    spec_trai_error = zeros(n_repe,1);
    spec_test_error = zeros(n_repe,1);
    
    for repetition = 1:n_repe
        neural_net = fitnet(n_hidden_nodes);
        neural_net.trainParam.showWindow = 0;
        neural_net.trainParam.min_grad = 1e-10;
        neural_net.trainParam.max_fail = 20;
        [neural_net training_record] = train(neural_net,Xt,yt);
        chosen_epoch = training_record.num_epochs;

        spec_vali_error(repetition) = training_record.vperf(chosen_epoch);
        spec_trai_error(repetition) = training_record.perf(chosen_epoch);
        spec_test_error(repetition) = training_record.tperf(chosen_epoch);
    end
    vali_error(n_experim) = median(spec_vali_error);
    trai_error(n_experim) = median(spec_trai_error);
    test_error(n_experim) = median(spec_test_error);
    fprintf('Iteration %i (%i hidden nodes)\n',n_experim,n_hidden_nodes);
    n_experim = n_experim + 1;
end

%% Result visualization
% Setting color for lines
testColor = [1 0 0];
trainColor = [0 0 1];
valColor = [0 0.8 0];

% Plotting performances for Training, Validation and Test sets
figure();
semilogy(min_hidden_nodes:max_hidden_nodes,trai_error,'Color',trainColor);
hold on;
semilogy(min_hidden_nodes:max_hidden_nodes,vali_error,'Color',valColor);
semilogy(min_hidden_nodes:max_hidden_nodes,test_error,'Color',testColor);

xlabel('Hidden unit number');
ylabel('MSE (mean squared error)');
legend({'Train' 'Validation' 'Test'});
